Package: hmmTensor
Type: Package
Title: Hidden Markov Model by Matrix and Tensor Decomposition
Version: 0.1.0
Authors@R: person(given = "Koki",
    family = "Tsuyuzaki",
    role = c("aut", "cre"),
    email = "k.t.the-answer@hotmail.co.jp")
Description: Solves Hidden Markov Models (HMMs) via matrix and tensor
    decomposition. Converts observation sequences to co-occurrence
    matrices/tensors and applies Symmetric Non-negative Matrix
    Factorization (symNMF), Singular Value Decomposition (SVD),
    CANDECOMP/PARAFAC (CP) decomposition, or Tensor-Train (TT)
    decomposition to recover HMM parameters.
    Also provides standard HMM algorithms (Forward, Backward, Viterbi,
    Baum-Welch) for comparison.
    The spectral learning approach for HMMs is based on
    Hsu, Kakade, and Zhang (2012) <doi:10.1016/j.jcss.2011.12.025>.
    The symNMF method is described in
    Kuang, Yun, and Park (2015) <doi:10.1007/s10898-014-0247-2>.
    The Tensor-Train decomposition is described in
    Oseledets (2011) <doi:10.1137/090752286>.
License: MIT + file LICENSE
Encoding: UTF-8
Depends: R (>= 3.5.0)
Imports: rTensor, symTensor, methods, stats
Suggests: testthat
RoxygenNote: 7.3.2
NeedsCompilation: no
Packaged: 2026-05-19 11:03:11 UTC; koki
Author: Koki Tsuyuzaki [aut, cre]
Maintainer: Koki Tsuyuzaki <k.t.the-answer@hotmail.co.jp>
Repository: CRAN
Date/Publication: 2026-05-27 09:10:02 UTC
Built: R 4.5.2; ; 2026-05-27 11:07:36 UTC; unix

Package: hmmTensor
Type: Package
Title: Hidden Markov Model by Matrix and Tensor Decomposition
Version: 0.1.0
Authors@R: person(given = "Koki",
    family = "Tsuyuzaki",
    role = c("aut", "cre"),
    email = "k.t.the-answer@hotmail.co.jp")
Description: Solves Hidden Markov Models (HMMs) via matrix and tensor
    decomposition. Converts observation sequences to co-occurrence
    matrices/tensors and applies Symmetric Non-negative Matrix
    Factorization (symNMF), Singular Value Decomposition (SVD),
    CANDECOMP/PARAFAC (CP) decomposition, or Tensor-Train (TT)
    decomposition to recover HMM parameters.
    Also provides standard HMM algorithms (Forward, Backward, Viterbi,
    Baum-Welch) for comparison.
    The spectral learning approach for HMMs is based on
    Hsu, Kakade, and Zhang (2012) <doi:10.1016/j.jcss.2011.12.025>.
    The symNMF method is described in
    Kuang, Yun, and Park (2015) <doi:10.1007/s10898-014-0247-2>.
    The Tensor-Train decomposition is described in
    Oseledets (2011) <doi:10.1137/090752286>.
License: MIT + file LICENSE
Encoding: UTF-8
Depends: R (>= 3.5.0)
Imports: rTensor, symTensor, methods, stats
Suggests: testthat
RoxygenNote: 7.3.2
NeedsCompilation: no
Packaged: 2026-05-19 11:03:11 UTC; koki
Author: Koki Tsuyuzaki [aut, cre]
Maintainer: Koki Tsuyuzaki <k.t.the-answer@hotmail.co.jp>
Repository: CRAN
Date/Publication: 2026-05-27 09:10:02 UTC
Built: R 4.6.0; ; 2026-05-27 11:56:30 UTC; unix

import(rTensor)
import(symTensor)
import(methods)
importFrom(stats, runif, rnorm, rmultinom, rgamma)

export(Forward)
export(Backward)
export(Viterbi)
export(BaumWelch)
export(Seq2Prob)
export(HMM)
export(toyModel)

#' Backward Algorithm for HMM
#'
#' Computes backward probabilities \eqn{\beta_t(i) = P(Y_{t+1}, \ldots, Y_T | X_t = i)}
#' using the same scaling factors from the Forward algorithm.
#'
#' @param Y Integer vector of observations (values in 1:N)
#' @param T_mat Transition matrix (K x K), \code{T_mat[i,j] = P(X_t=j | X_{t-1}=i)}
#' @param O Emission matrix (K x N), \code{O[i,j] = P(Y_t=j | X_t=i)}
#' @param scale Scaling factors from \code{\link{Forward}} (length T_len)
#' @return Matrix (K x T_len) of scaled backward probabilities
#' @export
Backward <- function(Y, T_mat, O, scale) {
  K <- nrow(T_mat)
  T_len <- length(Y)

  beta <- matrix(0, K, T_len)

  # Initialization: beta_T(i) = 1 (scaled)
  beta[, T_len] <- 1 / scale[T_len]

  # Recursion
  for (t in (T_len - 1):1) {
    beta[, t] <- T_mat %*% (O[, Y[t + 1]] * beta[, t + 1])
    beta[, t] <- beta[, t] / scale[t]
  }

  beta
}

library(hmmTensor)
library(rTensor)

context("hmmTensor")

test_that("toyModel generates valid data", {
  toy <- toyModel("simple", T_len = 200, seed = 42)
  expect_identical(toy$K, 2L)
  expect_identical(toy$N, 3L)
  expect_identical(length(toy$Y), 200L)
  expect_identical(length(toy$X), 200L)
  expect_identical(all(toy$Y %in% 1:3), TRUE)
  expect_identical(all(toy$X %in% 1:2), TRUE)
})

test_that("toyModel types work", {
  for (tp in c("simple", "weather", "leftright")) {
    toy <- toyModel(tp, T_len = 100, seed = 1)
    expect_gt(toy$K, 1L)
    expect_gt(toy$N, 1L)
  }
})

test_that("Forward computes valid probabilities", {
  toy <- toyModel("simple", T_len = 100, seed = 42)
  fwd <- Forward(toy$Y, toy$T_mat, toy$O, toy$pi0)
  expect_identical(nrow(fwd$alpha), toy$K)
  expect_identical(ncol(fwd$alpha), length(toy$Y))
  # Each column should sum to ~1 after scaling
  col_sums <- colSums(fwd$alpha)
  expect_lt(max(abs(col_sums - 1)), 1e-10)
  expect_identical(is.finite(fwd$loglik), TRUE)
})

test_that("Backward returns correct dimensions", {
  toy <- toyModel("simple", T_len = 100, seed = 42)
  fwd <- Forward(toy$Y, toy$T_mat, toy$O, toy$pi0)
  bwd <- Backward(toy$Y, toy$T_mat, toy$O, fwd$scale)
  expect_identical(nrow(bwd), toy$K)
  expect_identical(ncol(bwd), length(toy$Y))
})

test_that("Viterbi decoding is accurate", {
  toy <- toyModel("simple", T_len = 500, seed = 42)
  vit <- Viterbi(toy$Y, toy$T_mat, toy$O, toy$pi0)
  acc <- mean(vit$path == toy$X)
  expect_gt(acc, 0.7)
  expect_identical(length(vit$path), length(toy$Y))
})

test_that("BaumWelch recovers parameters", {
  toy <- toyModel("simple", T_len = 1000, seed = 42)
  bw <- BaumWelch(toy$Y, K = 2, N = 3, num.iter = 50)
  # Log-likelihood should increase
  expect_gt(tail(bw$loglik, 1), bw$loglik[1])
  # Rows of T and O should sum to 1
  expect_lt(max(abs(rowSums(bw$T_mat) - 1)), 1e-10)
  expect_lt(max(abs(rowSums(bw$O) - 1)), 1e-10)
})

test_that("Seq2Prob order=2 returns valid matrix", {
  Y <- c(1, 2, 3, 1, 2, 3, 1, 2, 3)
  P2 <- Seq2Prob(Y, N = 3, order = 2)
  expect_identical(dim(P2), c(3L, 3L))
  expect_lt(abs(sum(P2) - 1), 1e-10)
})

test_that("Seq2Prob order=3 returns valid tensor", {
  Y <- c(1, 2, 3, 1, 2, 3, 1, 2, 3)
  P3 <- Seq2Prob(Y, N = 3, order = 3)
  expect_identical(P3@modes, c(3L, 3L, 3L))
  expect_lt(abs(sum(P3@data) - 1), 1e-10)
})

test_that("HMM symNMF solver works", {
  toy <- toyModel("simple", T_len = 500, seed = 42)
  res <- HMM(toy$Y, K = 2, N = 3, solver = "symNMF", num.iter = 100)
  expect_identical(nrow(res$T_mat), 2L)
  expect_identical(ncol(res$O), 3L)
  expect_lt(max(abs(rowSums(res$T_mat) - 1)), 1e-10)
  expect_lt(max(abs(rowSums(res$O) - 1)), 1e-10)
})

test_that("HMM SVD solver works", {
  toy <- toyModel("simple", T_len = 500, seed = 42)
  res <- HMM(toy$Y, K = 2, N = 3, solver = "SVD")
  expect_identical(nrow(res$T_mat), 2L)
  expect_lt(max(abs(rowSums(res$T_mat) - 1)), 1e-10)
})

test_that("HMM CP solver works", {
  toy <- toyModel("simple", T_len = 500, seed = 42)
  res <- HMM(toy$Y, K = 2, N = 3, solver = "CP", num.iter = 25)
  expect_identical(nrow(res$T_mat), 2L)
  expect_lt(max(abs(rowSums(res$T_mat) - 1)), 1e-10)
})

test_that("HMM TT solver works", {
  toy <- toyModel("simple", T_len = 500, seed = 42)
  res <- HMM(toy$Y, K = 2, N = 3, solver = "TT")
  expect_identical(nrow(res$T_mat), 2L)
  expect_lt(max(abs(rowSums(res$T_mat) - 1)), 1e-10)
})

#' Convert Observation Sequences to Co-occurrence Matrix/Tensor
#'
#' Constructs empirical co-occurrence statistics from observation sequences.
#' These form the basis for matrix/tensor decomposition approaches to HMM.
#'
#' @param Y Integer vector of observations (values in 1:N), or a list of
#'   integer vectors for multiple sequences.
#' @param N Number of distinct observation symbols. If NULL, inferred from data.
#' @param order Co-occurrence order: 2 (pairwise, default) or 3 (triple).
#'   Order 2 gives an N x N matrix \eqn{P_{2,1}(i,j) = P(Y_2=i, Y_1=j)}.
#'   Order 3 gives an N x N x N tensor \eqn{P_{3}(i,j,k) = P(Y_3=i, Y_2=j, Y_1=k)}.
#' @param lag Lag for pairwise co-occurrence (default: 1).
#'   \eqn{\Omega^{(\tau)}(i,j) = P(Y_t=i, Y_{t+\tau}=j)}.
#'   Only used when \code{order = 2}.
#' @param smooth Laplace smoothing pseudo-count (default: 0).
#'   Adds \code{smooth} to all counts before normalization.
#' @return For \code{order = 2}: an N x N matrix (normalized).
#'   For \code{order = 3}: an rTensor Tensor object of dimension N x N x N.
#' @export
#' @examples
#' Y <- c(1, 2, 3, 1, 2, 3, 1, 2, 3)
#' P2 <- Seq2Prob(Y, N = 3, order = 2)
#' P3 <- Seq2Prob(Y, N = 3, order = 3)
Seq2Prob <- function(Y, N = NULL, order = 2L, lag = 1L, smooth = 0) {
  if (!is.list(Y)) Y <- list(Y)

  if (is.null(N)) {
    N <- max(unlist(Y))
  }

  if (order == 2L) {
    # Pairwise co-occurrence: P(Y_{t+lag} = i, Y_t = j)
    counts <- matrix(smooth, N, N)
    for (y in Y) {
      T_len <- length(y)
      if (T_len <= lag) next
      for (t in 1:(T_len - lag)) {
        counts[y[t + lag], y[t]] <- counts[y[t + lag], y[t]] + 1
      }
    }
    # Normalize to joint probability
    total <- sum(counts)
    if (total > 0) counts <- counts / total
    return(counts)

  } else if (order == 3L) {
    # Triple co-occurrence: P(Y_{t+2} = i, Y_{t+1} = j, Y_t = k)
    counts <- array(smooth, dim = c(N, N, N))
    for (y in Y) {
      T_len <- length(y)
      if (T_len < 3) next
      for (t in 1:(T_len - 2)) {
        counts[y[t + 2], y[t + 1], y[t]] <- counts[y[t + 2], y[t + 1], y[t]] + 1
      }
    }
    total <- sum(counts)
    if (total > 0) counts <- counts / total
    return(rTensor::as.tensor(counts))

  } else {
    stop("order must be 2 or 3")
  }
}

#' Baum-Welch Algorithm (EM) for HMM
#'
#' Estimates HMM parameters (T, O, pi) from observation sequences
#' using the Expectation-Maximization algorithm.
#'
#' @param Y Integer vector of observations (values in 1:N), or a list of
#'   integer vectors for multiple sequences.
#' @param K Number of hidden states
#' @param N Number of distinct observation symbols
#' @param initT Initial transition matrix (K x K). If NULL, random initialization.
#' @param initO Initial emission matrix (K x N). If NULL, random initialization.
#' @param initPi Initial state distribution (length K). If NULL, uniform.
#' @param num.iter Maximum EM iterations (default: 100)
#' @param thr Convergence threshold on log-likelihood relative change (default: 1e-6)
#' @param verbose Logical (default: FALSE)
#' @return A list with components:
#'   \describe{
#'     \item{T_mat}{Estimated transition matrix (K x K)}
#'     \item{O}{Estimated emission matrix (K x N)}
#'     \item{pi0}{Estimated initial distribution (length K)}
#'     \item{loglik}{Vector of log-likelihoods per iteration}
#'     \item{converged}{Logical}
#'     \item{iter}{Number of iterations}
#'   }
#' @export
BaumWelch <- function(Y, K, N,
                      initT = NULL, initO = NULL, initPi = NULL,
                      num.iter = 100L, thr = 1e-6,
                      verbose = FALSE) {
  eps <- .Machine$double.eps

  # Handle single sequence as list
  if (!is.list(Y)) Y <- list(Y)

  # Initialize parameters
  if (is.null(initT)) {
    T_mat <- matrix(runif(K * K), K, K)
    T_mat <- T_mat / rowSums(T_mat)
  } else {
    T_mat <- initT
  }

  if (is.null(initO)) {
    O <- matrix(runif(K * N), K, N)
    O <- O / rowSums(O)
  } else {
    O <- initO
  }

  if (is.null(initPi)) {
    pi0 <- rep(1 / K, K)
  } else {
    pi0 <- initPi
  }

  loglik_hist <- numeric(num.iter)
  converged <- FALSE

  for (iter in seq_len(num.iter)) {
    # Accumulators
    T_numer <- matrix(eps, K, K)
    T_denom <- rep(eps, K)
    O_numer <- matrix(eps, K, N)
    O_denom <- rep(eps, K)
    pi_accum <- rep(0, K)
    total_loglik <- 0

    for (seq_idx in seq_along(Y)) {
      y <- Y[[seq_idx]]
      T_len <- length(y)

      # E-step
      fwd <- Forward(y, T_mat, O, pi0)
      bwd <- Backward(y, T_mat, O, fwd$scale)
      total_loglik <- total_loglik + fwd$loglik

      # Posterior: gamma_t(i) = P(X_t = i | Y)
      gamma <- fwd$alpha * bwd
      # Correct scaling: alpha is scaled by 1/c_t, beta is scaled by 1/c_t
      # gamma needs normalization
      gamma_sum <- colSums(gamma)
      gamma_sum[gamma_sum == 0] <- eps
      gamma <- sweep(gamma, 2, gamma_sum, "/")

      # xi_t(i,j) = P(X_t=i, X_{t+1}=j | Y)
      for (t in 1:(T_len - 1)) {
        xi <- (fwd$alpha[, t] %o% (O[, y[t + 1]] * bwd[, t + 1])) * T_mat
        xi_sum <- sum(xi)
        if (xi_sum > 0) xi <- xi / xi_sum
        T_numer <- T_numer + xi
        T_denom <- T_denom + gamma[, t]
      }

      # Emission accumulation
      for (t in 1:T_len) {
        O_numer[, y[t]] <- O_numer[, y[t]] + gamma[, t]
        O_denom <- O_denom + gamma[, t]
      }

      pi_accum <- pi_accum + gamma[, 1]
    }

    loglik_hist[iter] <- total_loglik

    # Check convergence
    if (iter > 1) {
      rel_change <- abs(loglik_hist[iter] - loglik_hist[iter - 1]) /
        (abs(loglik_hist[iter]) + eps)
      if (rel_change < thr) {
        converged <- TRUE
        if (verbose) message("Converged at iteration ", iter)
        break
      }
    }

    if (verbose && (iter %% 10 == 0 || iter == 1)) {
      message("Iter ", iter, ": loglik = ", round(total_loglik, 4))
    }

    # M-step
    T_mat <- T_numer / T_denom
    T_mat <- T_mat / rowSums(T_mat)

    O <- O_numer / O_denom
    O <- O / rowSums(O)

    pi0 <- pi_accum / sum(pi_accum)
  }

  list(
    T_mat = T_mat,
    O = O,
    pi0 = pi0,
    loglik = loglik_hist[seq_len(iter)],
    converged = converged,
    iter = iter
  )
}

#' Viterbi Algorithm for HMM
#'
#' Finds the most likely state sequence given the observations
#' using the Viterbi algorithm (log-space).
#'
#' @param Y Integer vector of observations (values in 1:N)
#' @param T_mat Transition matrix (K x K), \code{T_mat[i,j] = P(X_t=j | X_{t-1}=i)}
#' @param O Emission matrix (K x N), \code{O[i,j] = P(Y_t=j | X_t=i)}
#' @param pi0 Initial state distribution (length K)
#' @return A list with components:
#'   \describe{
#'     \item{path}{Integer vector of most likely states (length T_len)}
#'     \item{loglik}{Log-likelihood of the best path}
#'   }
#' @export
Viterbi <- function(Y, T_mat, O, pi0) {
  K <- nrow(T_mat)
  T_len <- length(Y)
  eps <- .Machine$double.eps

  log_T <- log(T_mat + eps)
  log_O <- log(O + eps)
  log_pi <- log(pi0 + eps)

  # delta[i, t] = max log-prob of best path ending in state i at time t
  delta <- matrix(-Inf, K, T_len)
  psi <- matrix(0L, K, T_len)  # backpointer

  # Initialization
  delta[, 1] <- log_pi + log_O[, Y[1]]

  # Recursion
  for (t in 2:T_len) {
    for (j in 1:K) {
      candidates <- delta[, t - 1] + log_T[, j]
      psi[j, t] <- which.max(candidates)
      delta[j, t] <- candidates[psi[j, t]] + log_O[j, Y[t]]
    }
  }

  # Backtracking
  path <- integer(T_len)
  path[T_len] <- which.max(delta[, T_len])
  loglik <- delta[path[T_len], T_len]

  for (t in (T_len - 1):1) {
    path[t] <- psi[path[t + 1], t + 1]
  }

  list(path = path, loglik = loglik)
}

#' HMM Parameter Estimation via Matrix/Tensor Decomposition
#'
#' Estimates HMM parameters by decomposing the observation co-occurrence
#' matrix/tensor. Supports multiple decomposition solvers.
#'
#' The pipeline:
#' \enumerate{
#'   \item Convert observations to co-occurrence matrix \eqn{\Omega = P_{2,1}} via \code{\link{Seq2Prob}}
#'   \item Decompose \eqn{\Omega} using the chosen solver
#'   \item Recover HMM parameters (T, O, pi) from the decomposition
#' }
#'
#' For NMF-based methods, \eqn{\Omega \approx M \Theta M^\top} where
#' M relates to the emission matrix O and \eqn{\Theta} to \code{diag(pi) \%*\% T}.
#'
#' @param Y Integer vector of observations (values in 1:N), or a list of
#'   integer vectors for multiple sequences.
#' @param K Number of hidden states
#' @param N Number of distinct observation symbols. If NULL, inferred from data.
#' @param solver Decomposition method:
#'   \describe{
#'     \item{\code{"symNMF"}}{Symmetric NMF via \code{symTensor::symNMF} (default)}
#'     \item{\code{"SVD"}}{Truncated SVD of the co-occurrence matrix}
#'     \item{\code{"CP"}}{CP decomposition of the 3rd-order co-occurrence tensor via \code{rTensor::cp}}
#'     \item{\code{"TT"}}{Tensor-Train approximation (via Tucker + reshape)}
#'   }
#' @param Beta Beta-divergence parameter for symNMF (default: 2). Ignored for other solvers.
#' @param order Co-occurrence order for Seq2Prob: 2 (default) or 3.
#'   \code{order = 3} is required for \code{solver = "CP"}.
#' @param lag Lag for pairwise co-occurrence (default: 1)
#' @param smooth Laplace smoothing pseudo-count (default: 1e-10)
#' @param num.iter Maximum iterations for iterative solvers (default: 100)
#' @param thr Convergence threshold (default: 1e-10)
#' @param verbose Logical (default: FALSE)
#' @return A list with components:
#'   \describe{
#'     \item{T_mat}{Estimated transition matrix (K x K)}
#'     \item{O}{Estimated emission matrix (K x N)}
#'     \item{pi0}{Estimated initial distribution (length K)}
#'     \item{Omega}{Co-occurrence matrix/tensor used}
#'     \item{decomp}{Raw decomposition result}
#'     \item{solver}{Solver used}
#'     \item{RecError}{Reconstruction error (if available)}
#'   }
#' @export
#' @examples
#' set.seed(42)
#' toy <- toyModel(type = "simple")
#' result <- HMM(toy$Y, K = toy$K, N = toy$N, solver = "symNMF")
#' result$T_mat
HMM <- function(Y, K, N = NULL, solver = c("symNMF", "SVD", "CP", "TT"),
                Beta = 2, order = 2L, lag = 1L, smooth = 1e-10,
                num.iter = 100L, thr = 1e-10, verbose = FALSE) {
  solver <- match.arg(solver)
  eps <- .Machine$double.eps

  if (is.null(N)) {
    if (is.list(Y)) N <- max(unlist(Y)) else N <- max(Y)
  }

  # Force order=3 for CP
  if (solver == "CP" && order != 3L) {
    order <- 3L
    if (verbose) message("CP solver requires order=3, overriding.")
  }

  # Step 1: Build co-occurrence
  Omega <- Seq2Prob(Y, N = N, order = order, lag = lag, smooth = smooth)

  # Step 2: Decompose
  result <- switch(solver,
    "symNMF" = .solveSymNMF(Omega, K, Beta, num.iter, thr, verbose),
    "SVD"    = .solveSVD(Omega, K),
    "CP"     = .solveCP(Omega, K, num.iter),
    "TT"     = .solveTT(Omega, K, num.iter)
  )

  # Attach metadata
  result$Omega <- Omega
  result$solver <- solver
  result
}

#' @keywords internal
.solveSymNMF <- function(Omega, K, Beta, num.iter, thr, verbose) {
  # Omega (N x N) ≈ M S M^T
  # M -> emission-related, S -> diag(pi) * T related
  res <- symTensor::symNMF(Omega, K = K, model = "MSM", Beta = Beta,
                           num.iter = num.iter, thr = thr, verbose = verbose)
  M <- res$M
  S <- res$S

  # Recover O: normalize M row-wise (M is N x K)
  # O[k, n] = P(Y=n | X=k) -> transpose and normalize
  O_raw <- t(M)  # K x N
  O <- O_raw / rowSums(O_raw)
  O[O < .Machine$double.eps] <- .Machine$double.eps
  O <- O / rowSums(O)

  # Recover pi and T from S
  # S ≈ diag(pi) * T in the ideal case
  # pi_k proportional to sum of S[k, ]
  pi0 <- colSums(S)
  pi0 <- pi0 / sum(pi0)

  # T[i,j] = S[i,j] / pi_i
  T_mat <- sweep(S, 1, rowSums(S), "/")
  T_mat[T_mat < .Machine$double.eps] <- .Machine$double.eps
  T_mat <- T_mat / rowSums(T_mat)

  list(T_mat = T_mat, O = O, pi0 = pi0,
       decomp = res, RecError = tail(res$RecError, 1))
}

#' @keywords internal
.solveSVD <- function(Omega, K) {
  # SVD of P_{2,1}: Omega = U D V^T
  # Use top-K components
  sv <- svd(Omega, nu = K, nv = K)
  U <- sv$u[, 1:K, drop = FALSE]  # N x K
  D <- diag(sv$d[1:K], nrow = K)
  V <- sv$v[, 1:K, drop = FALSE]  # N x K

  # Observable operator approach:
  # O ≈ |U| normalized
  O_raw <- abs(t(U))  # K x N
  O <- O_raw / rowSums(O_raw)
  O[O < .Machine$double.eps] <- .Machine$double.eps
  O <- O / rowSums(O)

  # Recover T from the singular value structure
  # P_{2,1} = O^T diag(pi) T O -> U D V^T
  # Approximate: T ≈ D (as transition strength)
  T_mat <- abs(D)
  T_mat <- T_mat / rowSums(T_mat)

  pi0 <- sv$d[1:K]
  pi0 <- pi0 / sum(pi0)

  recon <- U %*% D %*% t(V)
  rec_err <- sqrt(sum((Omega - recon)^2))

  list(T_mat = T_mat, O = O, pi0 = pi0,
       decomp = sv, RecError = rec_err)
}

#' @keywords internal
.solveCP <- function(Omega, K, num.iter) {
  # Omega is 3rd-order tensor (N x N x N)
  # CP decomposition: T ≈ sum_r lambda_r * u_r o v_r o w_r
  cp_res <- rTensor::cp(Omega, num_components = K, max_iter = num.iter)

  # Factor matrices: U[[1]], U[[2]], U[[3]] each N x K
  U1 <- abs(cp_res$U[[1]])
  U2 <- abs(cp_res$U[[2]])
  U3 <- abs(cp_res$U[[3]])

  # For HMM 3-view: M1 (future), M2 (present), M3 (past)
  # O ≈ average of the three views, normalized
  O_raw <- t((U1 + U2 + U3) / 3)  # K x N
  O <- O_raw / rowSums(O_raw)
  O[O < .Machine$double.eps] <- .Machine$double.eps
  O <- O / rowSums(O)

  # Recover T and pi from lambdas and structure
  lambdas <- abs(cp_res$lambdas)
  pi0 <- lambdas / sum(lambdas)

  # Cross-mode structure gives transition info
  # T[i,j] from how mode-1 and mode-3 components relate
  T_mat <- t(U1) %*% U3  # K x K
  T_mat <- abs(T_mat)
  T_mat <- T_mat / rowSums(T_mat)

  rec_err <- cp_res$fnorm_resid

  list(T_mat = T_mat, O = O, pi0 = pi0,
       decomp = cp_res, RecError = rec_err)
}

#' @keywords internal
.solveTT <- function(Omega, K, num.iter) {
  # For order-2: use Tucker as approximation to TT
  # Tucker: Omega ≈ G x_1 U1 x_2 U2
  if (is.matrix(Omega)) {
    # 2D case: Tucker is equivalent to truncated SVD
    sv <- svd(Omega, nu = K, nv = K)
    U <- sv$u[, 1:K, drop = FALSE]
    D <- diag(sv$d[1:K], nrow = K)
    V <- sv$v[, 1:K, drop = FALSE]

    O_raw <- abs(t(U))
    O <- O_raw / rowSums(O_raw)
    O[O < .Machine$double.eps] <- .Machine$double.eps
    O <- O / rowSums(O)

    # Core as transition
    core <- t(U) %*% Omega %*% V  # K x K
    T_mat <- abs(core)
    T_mat <- T_mat / rowSums(T_mat)

    pi0 <- sv$d[1:K]
    pi0 <- pi0 / sum(pi0)

    recon <- U %*% D %*% t(V)
    rec_err <- sqrt(sum((Omega - recon)^2))

    list(T_mat = T_mat, O = O, pi0 = pi0,
         decomp = sv, RecError = rec_err)
  } else {
    # 3D tensor: Tucker decomposition
    tuck <- rTensor::tucker(Omega, ranks = c(K, K, K), max_iter = num.iter)

    U1 <- abs(tuck$U[[1]])
    O_raw <- t(U1)
    O <- O_raw / rowSums(O_raw)
    O[O < .Machine$double.eps] <- .Machine$double.eps
    O <- O / rowSums(O)

    # Core tensor -> transition
    core_mat <- matrix(tuck$Z@data, K, K * K)
    T_mat <- abs(core_mat %*% t(core_mat))
    T_mat <- T_mat / rowSums(T_mat)

    pi0 <- rep(1 / K, K)

    list(T_mat = T_mat, O = O, pi0 = pi0,
         decomp = tuck, RecError = tuck$fnorm_resid)
  }
}

#' Forward Algorithm for HMM
#'
#' Computes forward probabilities \eqn{\alpha_t(i) = P(Y_1, \ldots, Y_t, X_t = i)}
#' with log-scaling to avoid underflow.
#'
#' @param Y Integer vector of observations (values in 1:N)
#' @param T_mat Transition matrix (K x K), \code{T_mat[i,j] = P(X_t=j | X_{t-1}=i)}
#' @param O Emission matrix (K x N), \code{O[i,j] = P(Y_t=j | X_t=i)}
#' @param pi0 Initial state distribution (length K)
#' @return A list with components:
#'   \describe{
#'     \item{alpha}{Matrix (K x T_len) of scaled forward probabilities}
#'     \item{loglik}{Log-likelihood of the observation sequence}
#'     \item{scale}{Vector of scaling factors (length T_len)}
#'   }
#' @export
Forward <- function(Y, T_mat, O, pi0) {
  K <- nrow(T_mat)
  T_len <- length(Y)

  alpha <- matrix(0, K, T_len)
  scale <- numeric(T_len)

  # Initialization: alpha_1(i) = pi0(i) * O(i, Y_1)
  alpha[, 1] <- pi0 * O[, Y[1]]
  scale[1] <- sum(alpha[, 1])
  alpha[, 1] <- alpha[, 1] / scale[1]

  # Recursion
  for (t in 2:T_len) {
    alpha[, t] <- (t(T_mat) %*% alpha[, t - 1]) * O[, Y[t]]
    scale[t] <- sum(alpha[, t])
    alpha[, t] <- alpha[, t] / scale[t]
  }

  loglik <- sum(log(scale))

  list(alpha = alpha, loglik = loglik, scale = scale)
}

#' Generate Toy HMM Data
#'
#' Creates synthetic HMM data with visually clear structure for
#' demonstrations and testing.
#'
#' @param type Type of toy model:
#'   \describe{
#'     \item{\code{"simple"}}{2 states, 3 observations. States alternate with
#'       high probability. Clear emission separation.}
#'     \item{\code{"weather"}}{3 states (Sunny/Cloudy/Rainy), 4 observations
#'       (Walk/Shop/Clean/Stay). Classic weather HMM.}
#'     \item{\code{"leftright"}}{3 states with left-to-right transitions only.
#'       Models sequential processes.}
#'   }
#' @param T_len Length of observation sequence (default: 500)
#' @param seed Random seed (default: NULL)
#' @return A list with components:
#'   \describe{
#'     \item{Y}{Integer vector of observations}
#'     \item{X}{Integer vector of true hidden states}
#'     \item{T_mat}{True transition matrix}
#'     \item{O}{True emission matrix}
#'     \item{pi0}{True initial distribution}
#'     \item{K}{Number of hidden states}
#'     \item{N}{Number of observation symbols}
#'     \item{type}{Model type}
#'   }
#' @export
#' @examples
#' toy <- toyModel("simple", T_len = 200, seed = 42)
#' table(toy$X)
#' table(toy$Y)
toyModel <- function(type = c("simple", "weather", "leftright"),
                     T_len = 500L, seed = NULL) {
  type <- match.arg(type)
  if (!is.null(seed)) set.seed(seed)

  params <- switch(type,
    "simple" = list(
      K = 2L, N = 3L,
      T_mat = matrix(c(
        0.9, 0.1,
        0.2, 0.8
      ), 2, 2, byrow = TRUE),
      O = matrix(c(
        0.7, 0.2, 0.1,
        0.1, 0.2, 0.7
      ), 2, 3, byrow = TRUE),
      pi0 = c(0.6, 0.4)
    ),
    "weather" = list(
      K = 3L, N = 4L,
      T_mat = matrix(c(
        0.6, 0.3, 0.1,
        0.2, 0.5, 0.3,
        0.1, 0.3, 0.6
      ), 3, 3, byrow = TRUE),
      O = matrix(c(
        0.6, 0.2, 0.1, 0.1,
        0.1, 0.4, 0.3, 0.2,
        0.05, 0.1, 0.4, 0.45
      ), 3, 4, byrow = TRUE),
      pi0 = c(0.4, 0.3, 0.3)
    ),
    "leftright" = list(
      K = 3L, N = 4L,
      T_mat = matrix(c(
        0.7, 0.3, 0.0,
        0.0, 0.6, 0.4,
        0.0, 0.0, 1.0
      ), 3, 3, byrow = TRUE),
      O = matrix(c(
        0.7, 0.2, 0.1, 0.0,
        0.1, 0.5, 0.3, 0.1,
        0.0, 0.1, 0.3, 0.6
      ), 3, 4, byrow = TRUE),
      pi0 = c(1.0, 0.0, 0.0)
    )
  )

  # Generate sequences
  X <- integer(T_len)
  Y <- integer(T_len)

  X[1] <- sample(params$K, 1, prob = params$pi0)
  Y[1] <- sample(params$N, 1, prob = params$O[X[1], ])

  for (t in 2:T_len) {
    X[t] <- sample(params$K, 1, prob = params$T_mat[X[t - 1], ])
    Y[t] <- sample(params$N, 1, prob = params$O[X[t], ])
  }

  list(
    Y = Y,
    X = X,
    T_mat = params$T_mat,
    O = params$O,
    pi0 = params$pi0,
    K = params$K,
    N = params$N,
    type = type
  )
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Seq2Prob.R
\name{Seq2Prob}
\alias{Seq2Prob}
\title{Convert Observation Sequences to Co-occurrence Matrix/Tensor}
\usage{
Seq2Prob(Y, N = NULL, order = 2L, lag = 1L, smooth = 0)
}
\arguments{
\item{Y}{Integer vector of observations (values in 1:N), or a list of
integer vectors for multiple sequences.}

\item{N}{Number of distinct observation symbols. If NULL, inferred from data.}

\item{order}{Co-occurrence order: 2 (pairwise, default) or 3 (triple).
Order 2 gives an N x N matrix \eqn{P_{2,1}(i,j) = P(Y_2=i, Y_1=j)}.
Order 3 gives an N x N x N tensor \eqn{P_{3}(i,j,k) = P(Y_3=i, Y_2=j, Y_1=k)}.}

\item{lag}{Lag for pairwise co-occurrence (default: 1).
\eqn{\Omega^{(\tau)}(i,j) = P(Y_t=i, Y_{t+\tau}=j)}.
Only used when \code{order = 2}.}

\item{smooth}{Laplace smoothing pseudo-count (default: 0).
Adds \code{smooth} to all counts before normalization.}
}
\value{
For \code{order = 2}: an N x N matrix (normalized).
  For \code{order = 3}: an rTensor Tensor object of dimension N x N x N.
}
\description{
Constructs empirical co-occurrence statistics from observation sequences.
These form the basis for matrix/tensor decomposition approaches to HMM.
}
\examples{
Y <- c(1, 2, 3, 1, 2, 3, 1, 2, 3)
P2 <- Seq2Prob(Y, N = 3, order = 2)
P3 <- Seq2Prob(Y, N = 3, order = 3)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Viterbi.R
\name{Viterbi}
\alias{Viterbi}
\title{Viterbi Algorithm for HMM}
\usage{
Viterbi(Y, T_mat, O, pi0)
}
\arguments{
\item{Y}{Integer vector of observations (values in 1:N)}

\item{T_mat}{Transition matrix (K x K), \code{T_mat[i,j] = P(X_t=j | X_{t-1}=i)}}

\item{O}{Emission matrix (K x N), \code{O[i,j] = P(Y_t=j | X_t=i)}}

\item{pi0}{Initial state distribution (length K)}
}
\value{
A list with components:
  \describe{
    \item{path}{Integer vector of most likely states (length T_len)}
    \item{loglik}{Log-likelihood of the best path}
  }
}
\description{
Finds the most likely state sequence given the observations
using the Viterbi algorithm (log-space).
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/toyModel.R
\name{toyModel}
\alias{toyModel}
\title{Generate Toy HMM Data}
\usage{
toyModel(type = c("simple", "weather", "leftright"), T_len = 500L, seed = NULL)
}
\arguments{
\item{type}{Type of toy model:
\describe{
  \item{\code{"simple"}}{2 states, 3 observations. States alternate with
    high probability. Clear emission separation.}
  \item{\code{"weather"}}{3 states (Sunny/Cloudy/Rainy), 4 observations
    (Walk/Shop/Clean/Stay). Classic weather HMM.}
  \item{\code{"leftright"}}{3 states with left-to-right transitions only.
    Models sequential processes.}
}}

\item{T_len}{Length of observation sequence (default: 500)}

\item{seed}{Random seed (default: NULL)}
}
\value{
A list with components:
  \describe{
    \item{Y}{Integer vector of observations}
    \item{X}{Integer vector of true hidden states}
    \item{T_mat}{True transition matrix}
    \item{O}{True emission matrix}
    \item{pi0}{True initial distribution}
    \item{K}{Number of hidden states}
    \item{N}{Number of observation symbols}
    \item{type}{Model type}
  }
}
\description{
Creates synthetic HMM data with visually clear structure for
demonstrations and testing.
}
\examples{
toy <- toyModel("simple", T_len = 200, seed = 42)
table(toy$X)
table(toy$Y)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/BaumWelch.R
\name{BaumWelch}
\alias{BaumWelch}
\title{Baum-Welch Algorithm (EM) for HMM}
\usage{
BaumWelch(
  Y,
  K,
  N,
  initT = NULL,
  initO = NULL,
  initPi = NULL,
  num.iter = 100L,
  thr = 1e-06,
  verbose = FALSE
)
}
\arguments{
\item{Y}{Integer vector of observations (values in 1:N), or a list of
integer vectors for multiple sequences.}

\item{K}{Number of hidden states}

\item{N}{Number of distinct observation symbols}

\item{initT}{Initial transition matrix (K x K). If NULL, random initialization.}

\item{initO}{Initial emission matrix (K x N). If NULL, random initialization.}

\item{initPi}{Initial state distribution (length K). If NULL, uniform.}

\item{num.iter}{Maximum EM iterations (default: 100)}

\item{thr}{Convergence threshold on log-likelihood relative change (default: 1e-6)}

\item{verbose}{Logical (default: FALSE)}
}
\value{
A list with components:
  \describe{
    \item{T_mat}{Estimated transition matrix (K x K)}
    \item{O}{Estimated emission matrix (K x N)}
    \item{pi0}{Estimated initial distribution (length K)}
    \item{loglik}{Vector of log-likelihoods per iteration}
    \item{converged}{Logical}
    \item{iter}{Number of iterations}
  }
}
\description{
Estimates HMM parameters (T, O, pi) from observation sequences
using the Expectation-Maximization algorithm.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Forward.R
\name{Forward}
\alias{Forward}
\title{Forward Algorithm for HMM}
\usage{
Forward(Y, T_mat, O, pi0)
}
\arguments{
\item{Y}{Integer vector of observations (values in 1:N)}

\item{T_mat}{Transition matrix (K x K), \code{T_mat[i,j] = P(X_t=j | X_{t-1}=i)}}

\item{O}{Emission matrix (K x N), \code{O[i,j] = P(Y_t=j | X_t=i)}}

\item{pi0}{Initial state distribution (length K)}
}
\value{
A list with components:
  \describe{
    \item{alpha}{Matrix (K x T_len) of scaled forward probabilities}
    \item{loglik}{Log-likelihood of the observation sequence}
    \item{scale}{Vector of scaling factors (length T_len)}
  }
}
\description{
Computes forward probabilities \eqn{\alpha_t(i) = P(Y_1, \ldots, Y_t, X_t = i)}
with log-scaling to avoid underflow.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/HMM.R
\name{HMM}
\alias{HMM}
\title{HMM Parameter Estimation via Matrix/Tensor Decomposition}
\usage{
HMM(
  Y,
  K,
  N = NULL,
  solver = c("symNMF", "SVD", "CP", "TT"),
  Beta = 2,
  order = 2L,
  lag = 1L,
  smooth = 1e-10,
  num.iter = 100L,
  thr = 1e-10,
  verbose = FALSE
)
}
\arguments{
\item{Y}{Integer vector of observations (values in 1:N), or a list of
integer vectors for multiple sequences.}

\item{K}{Number of hidden states}

\item{N}{Number of distinct observation symbols. If NULL, inferred from data.}

\item{solver}{Decomposition method:
\describe{
  \item{\code{"symNMF"}}{Symmetric NMF via \code{symTensor::symNMF} (default)}
  \item{\code{"SVD"}}{Truncated SVD of the co-occurrence matrix}
  \item{\code{"CP"}}{CP decomposition of the 3rd-order co-occurrence tensor via \code{rTensor::cp}}
  \item{\code{"TT"}}{Tensor-Train approximation (via Tucker + reshape)}
}}

\item{Beta}{Beta-divergence parameter for symNMF (default: 2). Ignored for other solvers.}

\item{order}{Co-occurrence order for Seq2Prob: 2 (default) or 3.
\code{order = 3} is required for \code{solver = "CP"}.}

\item{lag}{Lag for pairwise co-occurrence (default: 1)}

\item{smooth}{Laplace smoothing pseudo-count (default: 1e-10)}

\item{num.iter}{Maximum iterations for iterative solvers (default: 100)}

\item{thr}{Convergence threshold (default: 1e-10)}

\item{verbose}{Logical (default: FALSE)}
}
\value{
A list with components:
  \describe{
    \item{T_mat}{Estimated transition matrix (K x K)}
    \item{O}{Estimated emission matrix (K x N)}
    \item{pi0}{Estimated initial distribution (length K)}
    \item{Omega}{Co-occurrence matrix/tensor used}
    \item{decomp}{Raw decomposition result}
    \item{solver}{Solver used}
    \item{RecError}{Reconstruction error (if available)}
  }
}
\description{
Estimates HMM parameters by decomposing the observation co-occurrence
matrix/tensor. Supports multiple decomposition solvers.
}
\details{
The pipeline:
\enumerate{
  \item Convert observations to co-occurrence matrix \eqn{\Omega = P_{2,1}} via \code{\link{Seq2Prob}}
  \item Decompose \eqn{\Omega} using the chosen solver
  \item Recover HMM parameters (T, O, pi) from the decomposition
}

For NMF-based methods, \eqn{\Omega \approx M \Theta M^\top} where
M relates to the emission matrix O and \eqn{\Theta} to \code{diag(pi) \%*\% T}.
}
\examples{
set.seed(42)
toy <- toyModel(type = "simple")
result <- HMM(toy$Y, K = toy$K, N = toy$N, solver = "symNMF")
result$T_mat
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Backward.R
\name{Backward}
\alias{Backward}
\title{Backward Algorithm for HMM}
\usage{
Backward(Y, T_mat, O, scale)
}
\arguments{
\item{Y}{Integer vector of observations (values in 1:N)}

\item{T_mat}{Transition matrix (K x K), \code{T_mat[i,j] = P(X_t=j | X_{t-1}=i)}}

\item{O}{Emission matrix (K x N), \code{O[i,j] = P(Y_t=j | X_t=i)}}

\item{scale}{Scaling factors from \code{\link{Forward}} (length T_len)}
}
\value{
Matrix (K x T_len) of scaled backward probabilities
}
\description{
Computes backward probabilities \eqn{\beta_t(i) = P(Y_{t+1}, \ldots, Y_T | X_t = i)}
using the same scaling factors from the Forward algorithm.
}
